import os
import numpy as np
import tensorflow as tf
import shutil
import json
import time
import cv2 as cv

from py2vision.recognition.selector import NeuralNetwork
from py2vision.models.models_manager import ModelManager
from py2vision.models.blocks.backbone_block import BackboneBlock
from py2vision.models.blocks.backbone_block import darknet53, darknet19_tiny
from py2vision.compute.yolov3_calculus import YoloV3Calculus
from py2vision.utils.label_utils import read_class_names
from py2vision.image_process.frame_decorator import Frame
from py2vision.image_process.resize_with_bbox import ResizeWithBBox
from py2vision.datasets_loader.yolov3_dataset_generator import YoloV3DatasetGenerator

class ObjectDetectorYoloV3(NeuralNetwork):
    """Made of an Yolo network model and a dataset generator.
    
    Args:
        mode_name: an string to naming the model.
        num_class: an integer with the numbers of classes in the model.
        input_shape: A tuple with dims shape (height, weight, channels). 
        version: it can be 'yolov3' or 'yolov3_tiny'.
        training: a boolean that change depending if you want to train the model
        gpu_name: a gpu name if it is None this class search automatically a gpu compatible.

    Attributes:
        model: A model instance.
        num_class: an integer with the numbers of classes in the model.
        version: it can be 'yolov3' or 'yolov3_tiny'.
        model_name: an string to naming the model.
        input_shape: A tuple with dims shape (height, weight, channels). 
        gpus: a list with all allowed gpus.
        conv_tensors: these are the ouput of build yolov3 without prediction layer.
    """
    def __init__(self, model_name, num_class, input_shape=[416, 416, 3], version="yolov3", training=False, gpu_name=None):
        super().__init__()
        self.model = None
        self.num_class = num_class
        self.version = version
        self.model_name = model_name
        self.input_shape = input_shape

        if gpu_name == None:
            self.gpus = tf.config.experimental.list_physical_devices('GPU')
            if len(self.gpus) > 0:
                print(f'GPUs {self.gpus}')
                try: tf.config.experimental.set_memory_growth(self.gpus[0], True)
                except RuntimeError: pass
        else: 
            try: tf.config.experimental.set_memory_growth(gpu_name, True)
            except RuntimeError: pass

        if self.version == "yolov3":
            backbone_net = BackboneBlock(darknet53())
            model_manager = ModelManager()
            self.conv_tensors = model_manager.build_yolov3(backbone_net, self.num_class)(np.asarray(self.input_shape))
        elif self.version == "yolov3_tiny":
            backbone_net = BackboneBlock(darknet19_tiny())
            model_manager = ModelManager()
            self.conv_tensors = model_manager.build_yolov3_tiny(backbone_net, self.num_class)(np.asarray(self.input_shape))
        else: 
            versions = ["yolov3", "yolov3_tiny"]
            raise ValueError("yolo_version just can be: {}".format(", ".join(versions)))
        
        self.model = self.build_model(self.conv_tensors, training)
    
    def build_model(self, conv_tensors, training):
        """Build the complete yolo model and return model instance.
        
        Args:
            conv_tensors: a tensor with convolutional layers of a yolo network without output  layers or prediction layers.
            training: a boolean that change network structure, if is true the last layers will be predict tensors otherwise it will be output tensors.

        Returns:
            A yolo model.
        """
        output_tensors = []
        input_layer = conv_tensors[-1]
        self.training = training
        compute = YoloV3Calculus()
        for i, conv_tensor in enumerate(conv_tensors[:-1]):
            pred_tensor = compute.decode(conv_tensor, self.num_class, i)
            if self.training: output_tensors.append(conv_tensor)
            output_tensors.append(pred_tensor)

        return tf.keras.Model(input_layer, output_tensors, name=self.model_name)
        
    def train(self, train_annotations_path, test_annotations_path, class_file_name, 
                checkpoint_path="checkpoints", use_checkpoint=False, warmup_epochs=2, 
                epochs=100, log_dir="logs", save_only_best_model=True, save_all_checkpoints=False, batch_size=4, lr_init=1e-4, lr_end=1e-6, 
                strides=[8, 16, 32],
                anchors=[[[10,  13], [16,   30], [33,   23]],
                        [[30,  61], [62,   45], [59,  119]],
                        [[116, 90], [156, 198], [373, 326]]],
                anchor_per_scale=3, max_bbox_per_scale=100):
        """Train an yolov3 network or yolov3 tiny.
        
        Args:
            train_annotations_path: a string corresponding to the folder where train annotations are located.
            test_annotations_path: a string corresponding to the folder where test annotations are located.
            class_file_name: a string corresponding to the classes file (a .txt file with a list of classes) is located.
            checkpoint_path: a string corresponding to the checkpoint file that is inside of a checkpoints folder.
            use_checkpoint: a boolean that controls if use chepoint before train 
            warmup_epochs: an hiperparameter that update learning rate like this paper https://arxiv.org/pdf/1812.01187.pdf&usg=ALkJrhglKOPDjNt6SHGbphTHyMcT0cuMJg
            epochs: Number of epochs to train.
            log_dir: a folder to save logs.
            save_only_best_model: if is true the model will be saved when best validation loss > total validation loss/total test elements, but if it isn't true model will be saved always.
            save_all_checkpoints: it is a boolean, if is true model will be saved in each epoch.
            batch_size: an integer with the size of batches in test and train datasets.
            lr_init: a float which is initial learning rate
            lr_end: a float which is final learning rate
            strides: a list with the strides in a yolo model.
            anchors: these are the yolo anchors sizes.
            anchor_per_scale: an integer with the number of anchor boxes per scale. 
            max_bbox_per_scale: nan integer with the number of bounding boxes per scale. 
        """
        training = True
        if self.training == False:
            self.model = self.build_model(self.conv_tensors, training)

        if os.path.exists(log_dir): shutil.rmtree(log_dir)
        self.writer = tf.summary.create_file_writer(log_dir)
        train_set = YoloV3DatasetGenerator(train_annotations_path, class_file_name, batch_size=batch_size, strides=strides, anchors=anchors, anchor_per_scale=anchor_per_scale, max_bbox_per_scale=max_bbox_per_scale)
        test_set = YoloV3DatasetGenerator(test_annotations_path, class_file_name, batch_size=batch_size, strides=strides, anchors=anchors, anchor_per_scale=anchor_per_scale, max_bbox_per_scale=max_bbox_per_scale)

        steps_per_epoch = len(train_set)
        self.global_steps = tf.Variable(1, trainable=False, dtype=tf.int64)
        self.warmup_steps = warmup_epochs * steps_per_epoch
        self.total_steps = epochs * steps_per_epoch

        if use_checkpoint:
            try:
                self.model.load_weights(checkpoint_path)
            except ValueError:
                assert Exception("Shapes are incompatible between model and weights")
        
        checkpoint_path_splited = os.path.split(checkpoint_path)
        if len(checkpoint_path_splited[0]) == 0: 
            try:
                os.mkdir("checkpoints")
            except OSError as error:
                print(error)
            checkpoint_folder = "checkpoints"
        else: 
            checkpoint_folder = checkpoint_path_splited[0]
        optimizer = tf.keras.optimizers.Adam()

        validate_writer = tf.summary.create_file_writer(log_dir)

        mAP_model = self.build_model(self.conv_tensors, training) # create second model to measure mAP

        best_val_loss = 1000 # should be large at start
        for epoch in range(epochs):
            for image_data, target in train_set:
                results = self.train_step(image_data, target, optimizer, lr_init, lr_end)
                cur_step = results[0]%steps_per_epoch
                print("epoch:{:2.0f} step:{:5.0f}/{}, lr:{:.6f}, giou_loss:{:7.2f}, conf_loss:{:7.2f}, prob_loss:{:7.2f}, total_loss:{:7.2f}"
                    .format(epoch, cur_step, steps_per_epoch, results[1], results[2], results[3], results[4], results[5]))

            if len(test_set) == 0:
                print("configure TEST options to validate model")
                self.model.save_weights(os.path.join(checkpoint_folder, self.model._name))
                continue
            
            count, giou_val, conf_val, prob_val, total_val = 0., 0, 0, 0, 0
            for image_data, target in test_set:
                results = self.validate_step(image_data, target)
                count += 1
                giou_val += results[0]
                conf_val += results[1]
                prob_val += results[2]
                total_val += results[3]
            # writing validate summary data
            with validate_writer.as_default():
                tf.summary.scalar("validate_loss/total_val", total_val/count, step=epoch)
                tf.summary.scalar("validate_loss/giou_val", giou_val/count, step=epoch)
                tf.summary.scalar("validate_loss/conf_val", conf_val/count, step=epoch)
                tf.summary.scalar("validate_loss/prob_val", prob_val/count, step=epoch)
            validate_writer.flush()
                
            print("\n\ngiou_val_loss:{:7.2f}, conf_val_loss:{:7.2f}, prob_val_loss:{:7.2f}, total_val_loss:{:7.2f}\n\n".
                format(giou_val/count, conf_val/count, prob_val/count, total_val/count))

            if save_all_checkpoints and not save_only_best_model:
                save_directory = os.path.join(checkpoint_folder, self.model._name+"_val_loss_{:7.2f}_epoch_{}.ckpt".format(total_val/count, epoch))
                self.model.save_weights(save_directory)
            if save_only_best_model and best_val_loss>total_val/count:
                save_directory = os.path.join(checkpoint_folder, self.model._name+"best_val_loss_{:7.2f}_epoch_{}.ckpt".format(total_val/count, epoch))
                self.model.save_weights(save_directory)
                best_val_loss = total_val/count
            if not save_only_best_model and not save_all_checkpoints:
                save_directory = os.path.join(checkpoint_folder, self.model._name+"_val_loss_{:7.2f}_epoch_{}.ckpt".format(total_val/count, epoch))
                self.model.save_weights(save_directory)
        
        # measure mAP of trained custom model
        if 'save_directory' in locals():
            mAP_model.load_weights(save_directory) # use keras weights
            self.evaluate(mAP_model, test_set, class_file_name, 0.3, 0.45)

    
    def train_step(self, image_data, target, optimizer, lr_init=1e-4, lr_end=1e-6):
        """ training step.
        
        Args: 
            image_data: an image.
            target: labels
            optimizer: an tensorflow optimizer like Adams optimizer.
            lr_init: initial leraning rate hiperparameter.
            lr_end: final learning rate hiperparameter.

        Returns:
            (global_steps, optimizer.lr, giou_loss, conf_loss, prob_loss, total_loss)
        """
        
        with tf.GradientTape() as tape:

            pred_result = self.model(image_data, training=True)
            giou_loss=conf_loss=prob_loss=0
            compute = YoloV3Calculus()
            # optimizing process
            grid = 3 if not (self.version == "yolov3_tiny") else 2
            for i in range(grid):
                conv, pred = pred_result[i*2], pred_result[i*2+1]
                loss_items = compute.loss(pred, conv, *target[i], self.num_class, i)
                giou_loss += loss_items[0]
                conf_loss += loss_items[1]
                prob_loss += loss_items[2]

            total_loss = giou_loss + conf_loss + prob_loss

            gradients = tape.gradient(total_loss, self.model.trainable_variables)
            optimizer.apply_gradients(zip(gradients, self.model.trainable_variables))

            # update learning rate
            # about warmup: https://arxiv.org/pdf/1812.01187.pdf&usg=ALkJrhglKOPDjNt6SHGbphTHyMcT0cuMJg
            self.global_steps.assign_add(1)
            if self.global_steps < self.warmup_steps:# and not TRAIN_TRANSFER:
                lr = self.global_steps / self.warmup_steps * lr_init
            else:
                lr = lr_end + 0.5 * (lr_init - lr_end)*(
                    (1 + tf.cos((self.global_steps - self.warmup_steps) / (self.total_steps - self.warmup_steps) * np.pi)))
            optimizer.lr.assign(lr.numpy())

            # writing summary data
            with self.writer.as_default():
                tf.summary.scalar("lr", optimizer.lr, step=self.global_steps)
                tf.summary.scalar("loss/total_loss", total_loss, step=self.global_steps)
                tf.summary.scalar("loss/giou_loss", giou_loss, step=self.global_steps)
                tf.summary.scalar("loss/conf_loss", conf_loss, step=self.global_steps)
                tf.summary.scalar("loss/prob_loss", prob_loss, step=self.global_steps)
            self.writer.flush()
            
        return self.global_steps.numpy(), optimizer.lr.numpy(), giou_loss.numpy(), conf_loss.numpy(), prob_loss.numpy(), total_loss.numpy()

    def validate_step(self, image_data, target):
        """ Validation step.

        Args: 
            image_data: an image.
            target: labels.

        Returns:
            (giou_loss, conf_loss, prob_loss, total_loss)
        """
        with tf.GradientTape() as tape:
            pred_result = self.model(image_data, training=False)
            giou_loss=conf_loss=prob_loss=0
            compute = YoloV3Calculus()
            # optimizing process
            grid = 3 if not (self.version == "yolov3_tiny") else 2
            for i in range(grid):
                conv, pred = pred_result[i*2], pred_result[i*2+1]
                loss_items = compute.loss(pred, conv, *target[i], self.num_class, i)
                giou_loss += loss_items[0]
                conf_loss += loss_items[1]
                prob_loss += loss_items[2]

            total_loss = giou_loss + conf_loss + prob_loss
            
        return giou_loss.numpy(), conf_loss.numpy(), prob_loss.numpy(), total_loss.numpy()

    def restore_weights(self, weights_file, use_checkpoint=False):
        """Load previously trained model weights.

        Args: 
            weights_file: beginning by project root this is the path where is save your weights; example: "weights/weights_01.h5".
            use_checkpoint: if you wanna use a .ckpt file this variable should be True.
        """
        tf.keras.backend.clear_session() # used to reset layer names
        # load Darknet original weights to TensorFlow model
        if use_checkpoint: 
            try:
                self.model.load_weights(weights_file)
            except ValueError:
                assert Exception("Shapes are incompatible between model and weights")
        else:
            range1 = 75 if self.version == "yolov3" else 13
            range2 = [58, 66, 74] if self.version == "yolov3" else [9, 12]

            with open(weights_file, 'rb') as wf:
                major, minor, revision, seen, _ = np.fromfile(wf, dtype=np.int32, count=5)

                j = 0
                for i in range(range1):
                    if i > 0:
                        conv_layer_name = 'conv2d_%d' %i
                    else:
                        conv_layer_name = 'conv2d'
                        
                    if j > 0:
                        bn_layer_name = 'batch_normalization_%d' %j
                    else:
                        bn_layer_name = 'batch_normalization'
                    
                    conv_layer = self.model.get_layer(conv_layer_name)
                    filters = conv_layer.filters
                    k_size = conv_layer.kernel_size[0]
                    in_dim = conv_layer.input_shape[-1]

                    if i not in range2:
                        # darknet weights: [beta, gamma, mean, variance]
                        bn_weights = np.fromfile(wf, dtype=np.float32, count=4 * filters)
                        # tf weights: [gamma, beta, mean, variance]
                        bn_weights = bn_weights.reshape((4, filters))[[1, 0, 2, 3]]
                        bn_layer = self.model.get_layer(bn_layer_name)
                        j += 1
                    else:
                        conv_bias = np.fromfile(wf, dtype=np.float32, count=filters)

                    # darknet shape (out_dim, in_dim, height, width)
                    conv_shape = (filters, in_dim, k_size, k_size)
                    conv_weights = np.fromfile(wf, dtype=np.float32, count=np.product(conv_shape))
                    # tf shape (height, width, in_dim, out_dim)
                    conv_weights = conv_weights.reshape(conv_shape).transpose([2, 3, 1, 0])

                    if i not in range2:
                        conv_layer.set_weights([conv_weights])
                        bn_layer.set_weights(bn_weights)
                    else:
                        conv_layer.set_weights([conv_weights, conv_bias])

                assert len(wf.read()) == 0, 'failed to read all data'


    def inference(self, image_path, input_size=416, score_threshold=0.3, iou_threshold=0.45, nms_method="nms"):
        """ Apply inference with trained model.

        Args:
            image_path: a path to an image.
            input_size: integer to resize an input image from their original dimensions to an square image.
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm
            nms_method: a string that can be  'nms' or 'soft-nms'.

        Returns:
            An array with bounding boxes
        """
        original_image = cv.imread(image_path)
        original_image = cv.cvtColor(original_image, cv.COLOR_BGR2RGB)
        original_image = cv.cvtColor(original_image, cv.COLOR_BGR2RGB)
        frame = Frame(np.copy(original_image))
        image_data = ResizeWithBBox(frame).apply([input_size, input_size])
        image_data = image_data[np.newaxis, ...].astype(np.float32)
        pred_bbox = self.model.predict(image_data)

        pred_bbox = [tf.reshape(x, (-1, tf.shape(x)[-1])) for x in pred_bbox]
        pred_bbox = tf.concat(pred_bbox, axis=0)
        compute = YoloV3Calculus()
        bboxes = compute.postprocess_boxes(pred_bbox, original_image, input_size, score_threshold)
        bboxes = compute.nms(bboxes, iou_threshold, method=nms_method)

        return bboxes


    def evaluate(self, model, dataset, classes_file, score_threshold=0.05, iou_threshold=0.50, test_input_size=416):
        """Apply evaluation using mAP.

        Args:
            model: a tesorflow detection model.
            dataset: an YoloV3DatasetGenerator instance with test dataset.
            classes_file: a string corresponding to the classes file (a .txt file with a list of classes) is located.
            score_threshold: if the score of a bounding boxes is less than score_threshold, it will be discard.
            iou_threshold: a parameter between (0, 1) which is used for nms algorithm.
            test_input_size: integer to resize an input image from their original dimensions to an square image.
            
        Returns:
            mAP score
        """
        min_overlap = 0.5 # default value (defined in the PASCAL VOC2012 challenge)
        num_class = read_class_names(classes_file)

        ground_truth_dir_path = 'mAP/ground-truth'
        if os.path.exists(ground_truth_dir_path): shutil.rmtree(ground_truth_dir_path)

        if not os.path.exists('mAP'): os.mkdir('mAP')
        os.mkdir(ground_truth_dir_path)

        print(f'\ncalculating mAP{int(iou_threshold*100)}...\n')

        gt_counter_per_class = {}
        for index in range(dataset.num_samples):
            ann_dataset = dataset.annotations[index]

            original_image, bbox_data_gt = dataset.parse_annotation(ann_dataset, True)

            if len(bbox_data_gt) == 0:
                bboxes_gt = []
                classes_gt = []
            else:
                bboxes_gt, classes_gt = bbox_data_gt[:, :4], bbox_data_gt[:, 4]
            ground_truth_path = os.path.join(ground_truth_dir_path, str(index) + '.txt')
            num_bbox_gt = len(bboxes_gt)

            bounding_boxes = []
            for i in range(num_bbox_gt):
                class_name = num_class[classes_gt[i]]
                xmin, ymin, xmax, ymax = list(map(str, bboxes_gt[i]))
                bbox = xmin + " " + ymin + " " + xmax + " " +ymax
                bounding_boxes.append({"class_name":class_name, "bbox":bbox, "used":False})

                # count that object
                if class_name in gt_counter_per_class:
                    gt_counter_per_class[class_name] += 1
                else:
                    # if class didn't exist yet
                    gt_counter_per_class[class_name] = 1
                bbox_mess = ' '.join([class_name, xmin, ymin, xmax, ymax]) + '\n'
            with open(f'{ground_truth_dir_path}/{str(index)}_ground_truth.json', 'w') as outfile:
                json.dump(bounding_boxes, outfile)

        gt_classes = list(gt_counter_per_class.keys())
        # sort the classes alphabetically
        gt_classes = sorted(gt_classes)
        n_classes = len(gt_classes)

        compute = YoloV3Calculus()
        times = []
        json_pred = [[] for i in range(n_classes)]
        for index in range(dataset.num_samples):
            ann_dataset = dataset.annotations[index]

            image_name = ann_dataset[0].split('/')[-1]
            original_image, bbox_data_gt = dataset.parse_annotation(ann_dataset, True)
            frame = Frame(original_image)
            image = ResizeWithBBox(frame).apply([test_input_size, test_input_size])
            image_data = image[np.newaxis, ...].astype(np.float32)

            t1 = time.time()
            pred_bbox = model.predict(image_data)
            t2 = time.time()
            
            times.append(t2-t1)
            
            pred_bbox = [tf.reshape(x, (-1, tf.shape(x)[-1])) for x in pred_bbox]
            pred_bbox = tf.concat(pred_bbox, axis=0)
            bboxes = compute.postprocess_boxes(pred_bbox, original_image, test_input_size, score_threshold)
            bboxes = compute.nms(bboxes, iou_threshold, method='nms')

            for bbox in bboxes:
                coor = np.array(bbox[:4], dtype=np.int32)
                score = bbox[4]
                class_ind = int(bbox[5])
                class_name = num_class[class_ind]
                score = '%.4f' % score
                xmin, ymin, xmax, ymax = list(map(str, coor))
                bbox = xmin + " " + ymin + " " + xmax + " " +ymax
                json_pred[gt_classes.index(class_name)].append({"confidence": str(score), "file_id": str(index), "bbox": str(bbox)})

        ms = sum(times)/len(times)*1000
        fps = 1000 / ms

        for class_name in gt_classes:
            json_pred[gt_classes.index(class_name)].sort(key=lambda x:float(x['confidence']), reverse=True)
            with open(f'{ground_truth_dir_path}/{class_name}_predictions.json', 'w') as outfile:
                json.dump(json_pred[gt_classes.index(class_name)], outfile)

        # Calculate the AP for each class
        sum_AP = 0.0
        ap_dictionary = {}
        # open file to store the results
        with open("mAP/results.txt", 'w') as results_file:
            results_file.write("# AP and precision/recall per class\n")
            count_true_positives = {}
            for class_index, class_name in enumerate(gt_classes):
                count_true_positives[class_name] = 0
                # Load predictions of that class
                predictions_file = f'{ground_truth_dir_path}/{class_name}_predictions.json'
                predictions_data = json.load(open(predictions_file))

                # Assign predictions to ground truth objects
                nd = len(predictions_data)
                tp = [0] * nd # creates an array of zeros of size nd
                fp = [0] * nd
                for idx, prediction in enumerate(predictions_data):
                    file_id = prediction["file_id"]
                    # assign prediction to ground truth object if any
                    #   open ground-truth with that file_id
                    gt_file = f'{ground_truth_dir_path}/{str(file_id)}_ground_truth.json'
                    ground_truth_data = json.load(open(gt_file))
                    ovmax = -1
                    gt_match = -1
                    # load prediction bounding-box
                    bb = [ float(x) for x in prediction["bbox"].split() ] # bounding box of prediction
                    for obj in ground_truth_data:
                        # look for a class_name match
                        if obj["class_name"] == class_name:
                            bbgt = [ float(x) for x in obj["bbox"].split() ] # bounding box of ground truth
                            bi = [max(bb[0],bbgt[0]), max(bb[1],bbgt[1]), min(bb[2],bbgt[2]), min(bb[3],bbgt[3])]
                            iw = bi[2] - bi[0] + 1
                            ih = bi[3] - bi[1] + 1
                            if iw > 0 and ih > 0:
                                # compute overlap (IoU) = area of intersection / area of union
                                ua = (bb[2] - bb[0] + 1) * (bb[3] - bb[1] + 1) + (bbgt[2] - bbgt[0]
                                                + 1) * (bbgt[3] - bbgt[1] + 1) - iw * ih
                                ov = iw * ih / ua
                                if ov > ovmax:
                                    ovmax = ov
                                    gt_match = obj

                    # assign prediction as true positive/don't care/false positive
                    if ovmax >= min_overlap:# if ovmax > minimum overlap
                        if not bool(gt_match["used"]):
                            # true positive
                            tp[idx] = 1
                            gt_match["used"] = True
                            count_true_positives[class_name] += 1
                            # update the ".json" file
                            with open(gt_file, 'w') as f:
                                f.write(json.dumps(ground_truth_data))
                        else:
                            # false positive (multiple detection)
                            fp[idx] = 1
                    else:
                        # false positive
                        fp[idx] = 1

                # compute precision/recall
                cumsum = 0
                for idx, val in enumerate(fp):
                    fp[idx] += cumsum
                    cumsum += val
                cumsum = 0
                for idx, val in enumerate(tp):
                    tp[idx] += cumsum
                    cumsum += val
                #print(tp)
                rec = tp[:]
                for idx, val in enumerate(tp):
                    rec[idx] = float(tp[idx]) / gt_counter_per_class[class_name]
                #print(rec)
                prec = tp[:]
                for idx, val in enumerate(tp):
                    prec[idx] = float(tp[idx]) / (fp[idx] + tp[idx])
                #print(prec)

                ap, mrec, mprec = self.__voc_ap(rec, prec)
                sum_AP += ap
                text = "{0:.3f}%".format(ap*100) + " = " + class_name + " AP  " #class_name + " AP = {0:.2f}%".format(ap*100)

                rounded_prec = [ '%.3f' % elem for elem in prec ]
                rounded_rec = [ '%.3f' % elem for elem in rec ]
                # Write to results.txt
                results_file.write(text + "\n Precision: " + str(rounded_prec) + "\n Recall   :" + str(rounded_rec) + "\n\n")

                print(text)
                ap_dictionary[class_name] = ap

            results_file.write("\n# mAP of all classes\n")
            mAP = sum_AP / n_classes

            text = "mAP = {:.3f}%, {:.2f} FPS".format(mAP*100, fps)
            results_file.write(text + "\n")
            print(text)
            
            return mAP*100

    def __voc_ap(self, rec, prec):
        """
        --- Official matlab code VOC2012---
        mrec=[0 ; rec ; 1];
        mpre=[0 ; prec ; 0];
        for i=numel(mpre)-1:-1:1
                mpre(i)=max(mpre(i),mpre(i+1));
        end
        i=find(mrec(2:end)~=mrec(1:end-1))+1;
        ap=sum((mrec(i)-mrec(i-1)).*mpre(i));
        """
        rec.insert(0, 0.0) # insert 0.0 at begining of list
        rec.append(1.0) # insert 1.0 at end of list
        mrec = rec[:]
        prec.insert(0, 0.0) # insert 0.0 at begining of list
        prec.append(0.0) # insert 0.0 at end of list
        mpre = prec[:]
        """
        This part makes the precision monotonically decreasing
            (goes from the end to the beginning)
            matlab:  for i=numel(mpre)-1:-1:1
                                    mpre(i)=max(mpre(i),mpre(i+1));
        """
        # matlab indexes start in 1 but python in 0, so I have to do:
        #   range(start=(len(mpre) - 2), end=0, step=-1)
        # also the python function range excludes the end, resulting in:
        #   range(start=(len(mpre) - 2), end=-1, step=-1)
        for i in range(len(mpre)-2, -1, -1):
            mpre[i] = max(mpre[i], mpre[i+1])
        """
        This part creates a list of indexes where the recall changes
            matlab:  i=find(mrec(2:end)~=mrec(1:end-1))+1;
        """
        i_list = []
        for i in range(1, len(mrec)):
            if mrec[i] != mrec[i-1]:
                i_list.append(i) # if it was matlab would be i + 1
        """
        The Average Precision (AP) is the area under the curve
            (numerical integration)
            matlab: ap=sum((mrec(i)-mrec(i-1)).*mpre(i));
        """
        ap = 0.0
        for i in i_list:
            ap += ((mrec[i]-mrec[i-1])*mpre[i])
        return ap, mrec, mpre

    def print_summary(self):
        """Print network summary for debugging purposes."""
        self.model.summary()



import tensorflow as tf

from tensorflow.keras import Input 

from py2vision.models.layers.conv2d_bn_leaky_relu_layer import conv2d_bn_leaky_relu_layer
from py2vision.models.layers.upsample_layer import UpsampleLayer


class BuildYoloV3(tf.keras.Model):
    """Build YoloV3 model given a backbone.
        
    Args:
        backbone: an object with a backbone network.
        num_class: an integer with the quantity of classes. 
        
    Returns:
        A list where the first one is used to predict large-sized objects, 
        the second one is used to predict medium-sized objects, the third one is used to small objects
        and the last one is the input shape returned.
    """
    def __init__(self, backbone, num_class):
        super().__init__()
        if not isinstance(num_class, int):
            raise ValueError('num_class has to be an integer')
        self.base_outputs = backbone
        self.num_class = num_class
        self.first_stack_filters = [(1, 1, 1024,  512), (3, 3,  512, 1024), (1, 1, 1024,  512), 
                                    (3, 3,  512, 1024), (1, 1, 1024,  512)]
        self.second_stack_filters = [(1, 1, 768, 256), (3, 3, 256, 512), (1, 1, 512, 256), 
                                    (3, 3, 256, 512), (1, 1, 512, 256)]
        self.third_stack_filters = [(1, 1, 384, 128), (3, 3, 128, 256), (1, 1, 256, 128),
                                    (3, 3, 128, 256), (1, 1, 256, 128)]
        
    def call(self, input_shape):
        if len(input_shape) != 3: raise ValueError("input shape should have a len == 3")
        input_shape = Input([input_shape[0], input_shape[1], input_shape[2]])
        try:
            route_1, route_2, x = self.base_outputs.build_model(input_shape)
        except ValueError:
            raise Exception('Backbone output shape mismatch with yolov3 input shape')

        for n, filters in enumerate(self.first_stack_filters):
            x = conv2d_bn_leaky_relu_layer(x, filters)
        
        conv_lobj_branch = conv2d_bn_leaky_relu_layer(x, (3, 3, 512, 1024))
        # conv_lbbox is used to predict large-sized objects , Shape = [None, 13, 13, 255] (if num_class == 80 like coco dataset) 
        conv_lbbox = conv2d_bn_leaky_relu_layer(conv_lobj_branch, (1, 1, 1024, 3*(self.num_class + 5)), activate=False, bn=False)
        
        x = conv2d_bn_leaky_relu_layer(x, (1, 1,  512,  256))
        # upsample here uses the nearest neighbor interpolation method, which has the advantage that the
        # upsampling process does not need to learn, thereby reducing the network parameter  
        x = UpsampleLayer()(x)
        x = tf.concat([x, route_2], axis=-1)

        for n, filters in enumerate(self.second_stack_filters):
            x = conv2d_bn_leaky_relu_layer(x, filters)

        conv_mobj_branch = conv2d_bn_leaky_relu_layer(x, (3, 3, 256, 512))
        # conv_mbbox is used to predict medium-sized objects, shape = [None, 26, 26, 255] (if num_class == 80 like coco dataset) 
        conv_mbbox = conv2d_bn_leaky_relu_layer(conv_mobj_branch, (1, 1, 512, 3*(self.num_class + 5)), activate=False, bn=False)
        
        x = conv2d_bn_leaky_relu_layer(x, (1, 1, 256, 128))
        x = UpsampleLayer()(x)
        x = tf.concat([x, route_1], axis=-1)

        for n, filters in enumerate(self.third_stack_filters):
            x = conv2d_bn_leaky_relu_layer(x, filters)
        
        conv_sobj_branch = conv2d_bn_leaky_relu_layer(x, (3, 3, 128, 256))
        # conv_sbbox is used to predict small size objects, shape = [None, 52, 52, 255] (if num_class == 80 like coco dataset) 
        conv_sbbox = conv2d_bn_leaky_relu_layer(conv_sobj_branch, (1, 1, 256, 3*(self.num_class +5)), activate=False, bn=False)

        return [conv_sbbox, conv_mbbox, conv_lbbox, input_shape]


        









"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py2vision", # Replace with your own username
    version="1.0.1",
    license='MIT',
    author="Guillermo Jose Raven Lusinche",
    author_email="guillermoraven96@gmail.com",
    description="A package to implement a stereo vision system trained with deep learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/corvus96/PyTwoVision",
    packages=setuptools.find_packages(),
    package_dir={'py2vision': 'py2vision'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    python_requires='>=3.6',
     # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'numpy == 1.21.5',
        'tensorflow == 2.8.0',
        'opencv-contrib-python==4.6.0.66',
        'wget == 3.2',
        'matplotlib==3.5.1',
        'pandas',
        'pyyaml', 
        'h5py'
    ]
)


import os
import glob
import numpy as np
import cv2 as cv 
import requests
import time

from py2vision.stereo.stereo_builder import StereoSystemBuilder
from py2vision.stereo.match_method import Matcher
from py2vision.input_output.camera import Camera
from py2vision.utils.draw import draw_lines
from py2vision.compute.error_compute import re_projection_error


class StandardStereo:
    """An emulation of a real world stereo system with his relevant parameters, like fundamental matrix, rotation matrix, translation matrix and esential matrix. 

    Attributes:
        camL: left camera instance.
        camR: right camera instance.
        fish_eye: A boolean if is true it will calibrate with cv2.fisheye.calibrate if no it'll use normal calibration, fish eye is recomended when cameras has an field of view > 160.
    """
    def __init__(self, cam_left: Camera, cam_right: Camera, fish_eye=True):
        if type(fish_eye) != bool:
            raise ValueError("fish_eye has to be a boolean") 
        self.camL = cam_left
        self.camR = cam_right
        self.fish_eye = fish_eye
    
    def take_dual_photos(self, num_photos=15, save_dir_left="images_left", save_dir_right="images_right", prefix_name="photo"):
        """ A simple way to take photos in console and save in a folder.
        
        Args:
            num_photos: number of photos to take, 15 by default.
            save_dir_left: directory name where the left photos will be saved.
            save_dir_right: directory name where the right photos will be saved.
            prefix_name: A prefix for the names of the photos.

        Raises:
            OSError: if folder exist.
        """
        # initial directory
        cwd = os.getcwd()
        # make directory
        try:
            os.mkdir(save_dir_left)
        except OSError as error:
            print(error)   
        try:
            os.mkdir(save_dir_right)
        except OSError as error:
            print(error)   


        if self.camL.type_source == 'other' or self.camL.type_source == 'webcam':
            capL = cv.VideoCapture(self.camL.source)
            if not capL.isOpened():
                print("Cannot open left camera")
                exit()
        if self.camR.type_source == 'other' or self.camR.type_source == 'webcam':
            capR = cv.VideoCapture(self.camR.source)
            if not capR.isOpened():
                print("Cannot open right camera")
                exit()

        for i in range(num_photos):
            print("Please press enter to take a photo")
            while True:
                if self.camL.type_source == 'stream' and self.camR.type_source == 'stream':
                    img_respL = requests.get(self.camL.source)
                    img_arrL = np.array(bytearray(img_respL.content), dtype=np.uint8)
                    frameL = cv.imdecode(img_arrL, -1)
                    img_respR = requests.get(self.camR.source)
                    img_arrR = np.array(bytearray(img_respR.content), dtype=np.uint8)
                    frameR = cv.imdecode(img_arrR, -1)
                    cv.namedWindow("streaming: {}".format(self.camL.id), cv.WINDOW_NORMAL)
                    cv.imshow("streaming: {}".format(self.camL.id), frameL)
                    cv.namedWindow("streaming: {}".format(self.camR.id), cv.WINDOW_NORMAL)
                    cv.imshow("streaming: {}".format(self.camR.id), frameR)
                elif (self.camL.type_source == 'other' or 'webcam') and self.camR.type_source == 'stream':
                    _, frameL = capL.read()
                    img_respR = requests.get(self.camR.source)
                    img_arrR = np.array(bytearray(img_respR.content), dtype=np.uint8)
                    frameR = cv.imdecode(img_arrR, -1)
                    cv.imshow("webcam: {}".format(self.camL.id), frameL)
                    cv.namedWindow("streaming: {}".format(self.camR.id), cv.WINDOW_NORMAL)
                    cv.imshow("streaming: {}".format(self.camR.id), frameR)
                elif (self.camR.type_source == 'other' or 'webcam') and self.camL.type_source == 'stream':
                    _, frameR = capR.read()
                    img_respL = requests.get(self.camL.source)
                    img_arrL = np.array(bytearray(img_respL.content), dtype=np.uint8)
                    frameL = cv.imdecode(img_arrL, -1)
                    cv.imshow("webcam: {}".format(self.camR.id), frameR)
                    cv.namedWindow("streaming: {}".format(self.camL.id), cv.WINDOW_NORMAL)
                    cv.imshow("streaming: {}".format(self.camL.id), frameL)
                else:
                    _, frameL = capL.read()
                    _, frameR = capR.read()
                    cv.imshow("webcam: {}".format(self.camL.id), frameL)
                    cv.imshow("webcam: {}".format(self.camR.id), frameR)
                input_key = cv.waitKey(1)
                
                if input_key == 32:
                    #space pressed
                    os.chdir(save_dir_left)
                    cv.imwrite('{}_{}.png'.format(prefix_name, i + 1), frameL)
                    print("saved as {}_{}.png".format(prefix_name, i + 1))
                    os.chdir(cwd)
                    os.chdir(save_dir_right)
                    cv.imwrite('{}_{}.png'.format(prefix_name, i + 1), frameR)
                    print("saved as {}_{}.png".format(prefix_name, i + 1))
                    os.chdir(cwd)
                    print("{} photos left".format(num_photos - i - 1))
                    break
                if input_key == 27:
                    #esc pressed
                    print("Escape hit, closing...")
                    break
            if input_key == 27:
                #esc pressed
                break
        cv.destroyAllWindows()
        # restore to initial directory
        os.chdir(cwd)

    def calibrate(self, images_left_path, images_right_path, pattern_type='chessboard', pattern_size=(8,5), show=True):
        """ Compute intrinsics and extrinsics parameters for two cameras at once.
        Note: if you want a good calibration your images in 'images_left_path' and 'images_right_path' have to follow a format for example: example/left_images/photo_1.jpg and example/right_images/photo_1.jpg the number is important because calibrate method order the pair of images with sorted method.
        
        Args:
            images_left_path: folder where is saved left calibration pattern photos.
            images_right_path: folder where is saved right calibration pattern photos.
            pattern_type: It can be "circles" pattern or "chessboard" pattern (default).
            pattern_size: If pattern_type is "chessboard"  this the Number of inner corners per a chessboard row and column. But If pattern_type is "circles" this will be the number of circles per row and column. 
            show: if is true it show corners or centers found by calibration algorithm  at each iteration.
        
        Returns: 
            Three floats, The first is the reprojection error in left side,  the another one is right side
            and the third one is rms error (these have to be less than 1).

        Raises:
            OSError: If didn't find photos on images_left_path or images_right_path folders.
        """
        valid_patterns = ["chessboard", "circles"]
        if pattern_type not in valid_patterns:
            raise ValueError("pattern_type only can be 'chessboard' or 'circles'") 
        # adjust image path
        images_left, _ = os.path.splitext(images_left_path)
        images_right, _ = os.path.splitext(images_right_path)
        # available formats
        formats = ['*.svg', '*.png', '*.jpg', '*.bmp', '*.jpeg', '*.raw']
        # get files names
        filesL = [glob.glob((images_left + '/' + e)  if len(images_left) > 0 else e) for e in formats]
        filesR = [glob.glob((images_right + '/' + e)  if len(images_right) > 0 else e) for e in formats]
        # flatting files list
        imagesL = [item for sublist in filesL for item in sublist]
        imagesR = [item for sublist in filesR for item in sublist]
        imagesL = sorted(imagesL)
        imagesR = sorted(imagesR)
        # termination criteria
        if self.fish_eye == False:
            criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 30, 0.001)
        else:
            criteria = (cv.TERM_CRITERIA_EPS+cv.TERM_CRITERIA_MAX_ITER, 30, 0.1)
        
        # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
        if self.fish_eye == False:
            objp = np.zeros((pattern_size[1]*pattern_size[0],3), np.float32)
            objp[:,:2] = np.mgrid[0:pattern_size[0],0:pattern_size[1]].T.reshape(-1,2)
        else:
            objp = np.zeros((pattern_size[1]*pattern_size[0], 1, 3), np.float32)
            objp[:,0, :2] = np.mgrid[0:pattern_size[0],0:pattern_size[1]].T.reshape(-1,2)
        # Arrays to store object points and image points from all the images.
        objpoints = [] # 3d point in real world space
        imgpointsL = [] # 2d points in image plane.
        imgpointsR = [] # 2d points in image plane.
        try:
            if len(imagesL) == 0 or len(imagesR) == 0:
                raise OSError
            for imgLeft, imgRight in zip(imagesL, imagesR):
                imgL = cv.imread(imgLeft)
                imgR = cv.imread(imgRight)
                grayL = cv.cvtColor(imgL, cv.COLOR_BGR2GRAY)
                grayR = cv.cvtColor(imgR, cv.COLOR_BGR2GRAY)

                # Find the chess board corners
                if pattern_type == 'chessboard' and self.fish_eye == False:
                    retL, cornersL = cv.findChessboardCorners(grayL, pattern_size, None)
                    retR, cornersR = cv.findChessboardCorners(grayR, pattern_size, None)
                elif pattern_type == 'circles' and self.fish_eye == False:
                    retL, cornersL = cv.findCirclesGrid(grayL, pattern_size, None)
                    retR, cornersR = cv.findCirclesGrid(grayR, pattern_size, None)
                elif pattern_type == 'chessboard' and self.fish_eye == True:
                    retL, cornersL = cv.findChessboardCorners(grayL, pattern_size, flags=cv.CALIB_CB_ADAPTIVE_THRESH+cv.CALIB_CB_FAST_CHECK+cv.CALIB_CB_NORMALIZE_IMAGE)
                    retR, cornersR = cv.findChessboardCorners(grayR, pattern_size, flags=cv.CALIB_CB_ADAPTIVE_THRESH+cv.CALIB_CB_FAST_CHECK+cv.CALIB_CB_NORMALIZE_IMAGE)
                elif pattern_type == 'circles' and self.fish_eye == True:
                    retL, cornersL = cv.findCirclesGrid(grayL, pattern_size, cv.CALIB_CB_ADAPTIVE_THRESH+cv.CALIB_CB_FAST_CHECK+cv.CALIB_CB_NORMALIZE_IMAGE)
                    retR, cornersR = cv.findCirclesGrid(grayR, pattern_size, cv.CALIB_CB_ADAPTIVE_THRESH+cv.CALIB_CB_FAST_CHECK+cv.CALIB_CB_NORMALIZE_IMAGE)

                # If found, add object points, image points (after refining them)
                if retL and retR == True:

                    objpoints.append(objp)

                    if self.fish_eye == False:   
                        cornersL = cv.cornerSubPix(grayL, cornersL, (11,11), (-1,-1), criteria)
                        cornersR = cv.cornerSubPix(grayR, cornersR, (11,11), (-1,-1), criteria)
                    else:
                        cornersL = cv.cornerSubPix(grayL, cornersL, (3,3), (-1,-1), criteria)
                        cornersR = cv.cornerSubPix(grayR, cornersR, (3,3), (-1,-1), criteria)
                    imgpointsL.append(cornersL)
                    imgpointsR.append(cornersR)

                if show:
                    # Draw and display the corners
                    cv.drawChessboardCorners(imgL, pattern_size, cornersL, retL)
                    cv.namedWindow(imgLeft, cv.WINDOW_NORMAL)
                    cv.imshow(imgLeft, imgL)
                    cv.drawChessboardCorners(imgR, pattern_size, cornersR, retR)
                    cv.namedWindow(imgRight, cv.WINDOW_NORMAL)
                    cv.imshow(imgRight, imgR)
                    cv.waitKey(6000)
                    cv.destroyAllWindows()
            
            if self.fish_eye == False:
                flags = 0
                flags |= cv.CALIB_FIX_INTRINSIC
                # Here we fix the intrinsic camara matrixes so that only Rot, Trns, Emat and Fmat are calculated.
                # Hence intrinsic parameters are the same 

                criteria_stereo = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 30, 0.001)
            else:
                flags = cv.fisheye.CALIB_RECOMPUTE_EXTRINSIC + cv.fisheye.CALIB_FIX_SKEW
                criteria_stereo = (cv.TERM_CRITERIA_EPS+cv.TERM_CRITERIA_MAX_ITER, 30, 0.01)
                k = np.zeros((3, 3))
                d = np.zeros((4, 1))
                rvecs = [np.zeros((1, 1, 3), dtype=np.float64) for i in range(len(objpoints))]
                tvecs = [np.zeros((1, 1, 3), dtype=np.float64) for i in range(len(objpoints))]
            
            try:
                # This step is performed to transformation between the two cameras and calculate Essential and Fundamental matrix
                print("Calibrating estereo...")
                if self.fish_eye == False:
                    rms_error, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, self.rot, self.trans, self.e_matrix, self.f_matrix, _ = cv.stereoCalibrateExtended(objpoints, imgpointsL, imgpointsR, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, grayL.shape[::-1], criteria_stereo, flags)
                else:
                    rms_error, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, self.rot, self.trans, = cv.fisheye.stereoCalibrate(objpoints, imgpointsL, imgpointsR, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, grayL.shape[::-1], None, None, cv.CALIB_FIX_INTRINSIC, criteria_stereo)
                errorL = re_projection_error(objpoints, self.camL.rvecs, self.camL.tvecs, self.camL.matrix, imgpointsL, self.camL.dist_coeffs)
                print("Left camera error {}".format(errorL))
                errorR = re_projection_error(objpoints, self.camR.rvecs, self.camR.tvecs, self.camR.matrix, imgpointsR, self.camR.dist_coeffs)
                print("Right camera error {}".format(errorR))
                print("system calibrated")
                
            except AttributeError:
                print("Camera parameters hasn't fixed")
                print("Fixing left camera...")
                if self.fish_eye == False:
                    errorL, self.camL.matrix, self.camL.dist_coeffs, self.camL.rvecs, self.camL.tvecs = cv.calibrateCamera(objpoints, imgpointsL, grayL.shape[::-1], None, None)
                    heightL, widthL = imgL.shape[:2]
                    self.camL.matrix, self.camL.roi = cv.getOptimalNewCameraMatrix(self.camL.matrix, self.camL.dist_coeffs, (widthL, heightL), 1, (widthL, heightL))
                else:
                    errorL, self.camL.matrix, self.camL.dist_coeffs, self.camL.rvecs, self.camL.tvecs = cv.fisheye.calibrate(objpoints, imgpointsL, grayL.shape[::-1], k, d, rvecs, tvecs, flags, (cv.TERM_CRITERIA_EPS+cv.TERM_CRITERIA_MAX_ITER, 30, 1e-6))
                print("Left camera error {}".format(errorL))
                print("Fixing Right camera...")
                if  self.fish_eye == False:
                    errorR, self.camR.matrix, self.camR.dist_coeffs, self.camR.rvecs, self.camR.tvecs = cv.calibrateCamera(objpoints, imgpointsR, grayR.shape[::-1], None, None)
                    heightR, widthR = imgL.shape[:2]
                    self.camR.matrix, self.camR.roi = cv.getOptimalNewCameraMatrix(self.camR.matrix, self.camR.dist_coeffs, (widthR, heightR), 1, (widthR, heightR))
                else:
                    errorR, self.camR.matrix, self.camR.dist_coeffs, self.camR.rvecs, self.camR.tvecs = cv.fisheye.calibrate(objpoints, imgpointsR, grayR.shape[::-1], k, d, rvecs, tvecs, flags, (cv.TERM_CRITERIA_EPS+cv.TERM_CRITERIA_MAX_ITER, 30, 1e-6))
                print("Right camera error {}".format(errorR))
                print("Calibrating estereo...")
                # This step is performed to transformation between the two cameras and calculate Essential and Fundamenatl matrix
                if self.fish_eye == False:
                    self.ret_stereo, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, self.rot, self.trans, self.e_matrix, self.f_matrix, rms_error = cv.stereoCalibrateExtended(objpoints, imgpointsL, imgpointsR, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, grayL.shape[::-1], criteria_stereo, flags)
                else:
                    rms_error, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, self.rot, self.trans, = cv.fisheye.stereoCalibrate(objpoints, imgpointsL, imgpointsR, self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, grayL.shape[::-1], None, None, cv.CALIB_FIX_INTRINSIC, criteria_stereo)
                print("system calibrated")
        except OSError:
            if len(imagesL) == 0:
                print("Could not find any image in {}".format(images_left_path))
            if len(imagesR) == 0:
                print("Could not find any image in {}".format(images_right_path))

        return errorL, errorR, rms_error

    def rectify(self, image_left__dims, image_right_dims, export_file_name="stereoMap", alpha=1, output_size=(0,0), export_file=True):
        """ Compute stereo rectification maps and export left and right stereo maps in xml format. Note: you will need to calibrate first
        
        Args: 
            image_left__dims: a tuple or list with left image (width, height).
            image_right_dims: a tuple or list with right image (width, height).
            export_file_name: personalize the name of output file.
            alpha: free scaling parameter. If it is -1 or absent, the function performs the default scaling. Otherwise, the parameter should be between 0 and 1. alpha=0 means that the rectified images are zoomed and shifted so that only valid pixels are visible (no black areas after rectification). alpha=1 (default) means that the rectified image is decimated and shifted so that all the pixels from the original images from the cameras are retained in the rectified images (no source image pixels are lost). Any intermediate value yields an intermediate result between those two extreme cases (Only apply for no fish eye cameras). 
            output_size: New image resolution after rectification. When (0,0) is passed (default), it is set to the original image Size. Setting it to a larger value can help you preserve details in the original image, especially when there is a big radial distortion.
            export_file: if is true this method will save the parameters in an xml with the name in export_file_name. 
        
        Raises: 
            AttributeError: If haven't calibrated before, you need all parameters of calibrate() method.
         """
        if len(image_left__dims) != 2 or len(image_right_dims) != 2:
            raise ValueError("image_left_dims and image_right_dims must have a length of 2")
        if self.fish_eye:
            r2 = np.zeros([3,3])
            p1 = np.zeros([3,4])
            p2 = np.zeros([3,4])
            q = np.zeros([4,4])
        try:
            if self.fish_eye == False:
                rectL, rectR, projMatrixL, projMatrixR, self.Q, self.camL.roi, self.camR.roi = cv.stereoRectify(self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, image_left__dims, self.rot, self.trans, alpha, output_size, flags=cv.CALIB_ZERO_DISPARITY)

                self.stereoMapL = cv.initUndistortRectifyMap(self.camL.matrix, self.camL.dist_coeffs, rectL, projMatrixL, image_left__dims, cv.CV_16SC2)
                self.stereoMapR = cv.initUndistortRectifyMap(self.camR.matrix, self.camR.dist_coeffs, rectR, projMatrixR, image_right_dims, cv.CV_16SC2)
            else:
                rectL, rectR, projMatrixL, projMatrixR, self.Q = cv.fisheye.stereoRectify(self.camL.matrix, self.camL.dist_coeffs, self.camR.matrix, self.camR.dist_coeffs, image_left__dims, self.rot, self.trans, 0, r2, p1, p2, q, cv.CALIB_ZERO_DISPARITY, output_size, 0, 0)

                self.stereoMapL = cv.fisheye.initUndistortRectifyMap(self.camL.matrix, self.camL.dist_coeffs, rectL, projMatrixL, image_left__dims, cv.CV_16SC2)
                self.stereoMapR = cv.fisheye.initUndistortRectifyMap(self.camR.matrix, self.camR.dist_coeffs, rectR, projMatrixR, image_right_dims, cv.CV_16SC2)
            if export_file:
                print("Saving parameters!")
                cv_file = cv.FileStorage('{}.xml'.format(export_file_name), cv.FILE_STORAGE_WRITE)

                cv_file.write('stereoMapL_x', self.stereoMapL[0])
                cv_file.write('stereoMapL_y', self.stereoMapL[1])
                cv_file.write('stereoMapR_x', self.stereoMapR[0])
                cv_file.write('stereoMapR_y', self.stereoMapR[1])

                cv_file.release()
        except AttributeError:
            print("Please calibrate first!")
    
    def get_stereo_maps(self, path):
        """ Loads stereo maps parameters from a xml file
        
        Args: 
            path: path where is saved xml file. (Note: if you don't have stereo maps you can use calibrate method first, then use rectify method and pass true in export file argument)
         """
        # file storage read
        cv_file = cv.FileStorage(path, cv.FILE_STORAGE_READ)
        if not cv_file.isOpened():
            raise OSError("Looks like {} doesn't exist!".format(path))
        temp_stereoMapL = []
        temp_stereoMapR = []
        temp_stereoMapL.append(cv_file.getNode('stereoMapL_x').mat())
        temp_stereoMapL.append(cv_file.getNode('stereoMapL_y').mat())
        temp_stereoMapR.append(cv_file.getNode('stereoMapR_x').mat())
        temp_stereoMapR.append(cv_file.getNode('stereoMapR_y').mat())
        # copy to class attributes
        self.stereoMapL = temp_stereoMapL
        self.stereoMapR = temp_stereoMapR
        cv_file.release()
    
    def print_parameters(self, one_camera_parameters=['matrix', 'dist_coeffs'], stereo_parameters=['Q', 'rot', 'trans', 'e_matrix', 'f_matrix']):
        """ Print required individual camera parameters and stereo parameters already defined.

        Args: 
            one_camera_parameters: All elements in list  will be printed, they match with cameras parameters.
            stereo_parameters: All elements in list  will be printed, they match with stereo parameters. 
         """
        np.set_printoptions(suppress=True)
        print("Left camera:")
        for attr in dir(self.camL):
            if attr in one_camera_parameters:
                print("{} : {}".format(attr, getattr(self.camL, attr)))
        
        print("\n")
        print("Right camera:")
        for attr in dir(self.camR):
            if attr in one_camera_parameters:
                print("{} : {}".format(attr, getattr(self.camR, attr)))
        print("\n")
        print("Stereo:")
        for attr in dir(self):
            if attr in stereo_parameters:
                print("{} : {}".format(attr, getattr(self, attr)))
        np.set_printoptions(suppress=False)

class StandardStereoBuilder(StereoSystemBuilder):
    """
    Implement methods to get depth using OpenCV matchers like SGBM or BM.
    
    Attributes:
        camL: left camera instance.
        camR: right camera instance.
        stereo_maps_path: an xml with Stereo rectification maps
    """

    def __init__(self, cam_left: Camera, cam_right: Camera, stereo_maps_path) -> None:
        """
        A fresh builder instance should contain a blank product object, which is used in further assembly.
        """
        self.reset(cam_left, cam_right, stereo_maps_path)

    def reset(self, cam_left: Camera, cam_right: Camera, stereo_maps_path) -> None:
        """
        Initialitation of product
        """
        self._product = StandardStereo(cam_left, cam_right)
        self._product.get_stereo_maps(stereo_maps_path)

    def get_product(self):
        """
        Return: StandardStereo builded object
        """
        return self._product

    def pre_process(self, frameL, frameR, downsample=2):
        """ First, it transform from BGR to gray, next apply rectification and finally apply pyramid subsampling.
        
        Args: 
            frameL: it's the left frame
            frameR: it's the right frame
            downsample: if it is true, it will apply blurry in both frames and downsamples it. The downsampling factor just can be 2, 4, 8, 16, 32, 64. If downsample factor is 1 or None or False won't apply downsampling.
        
        Returns:
            Both frames to apply stereo correspondence or apply more processing to the images.
        """
        if downsample not in (2**p for p in range(1, 7)):
            if downsample in [1, None, False]:
                n_downsamples = 0
            else:
                raise ValueError("downsample only can be 2, 4, 8, 16, 32, 64")
        else:
            n_downsamples = [2**p for p in range(1, 7)].index(downsample)
            n_downsamples += 1
        
        img_left_gray = cv.cvtColor(frameL,cv.COLOR_BGR2GRAY)
        img_right_gray = cv.cvtColor(frameR,cv.COLOR_BGR2GRAY)

        # Applying stereo image rectification on the left image
        rectified_left = cv.remap(img_left_gray, self._product.stereoMapL[0], self._product.stereoMapL[1], cv.INTER_LANCZOS4, cv.BORDER_CONSTANT, 0)
        # Applying stereo image rectification on the right image
        rectified_right = cv.remap(img_right_gray, self._product.stereoMapR[0], self._product.stereoMapR[1], cv.INTER_LANCZOS4, cv.BORDER_CONSTANT, 0)
        if n_downsamples > 0:
            for i in range(n_downsamples):
                rectified_left = cv.pyrDown(rectified_left)
                rectified_right = cv.pyrDown(rectified_right)
        
        return rectified_left, rectified_right
    
    def find_epilines(self, frameL, frameR):
        """ Draw epilines to both frames.
        
        Args: 
            frameL: it's the left frame.
            frameR: it's the right frame.
        
        Returns:
            Two elements, the left and right frame with epilines
        """
        sift = cv.SIFT_create()
        # find the keypoints and descriptors with SIFT
        kp1, des1 = sift.detectAndCompute(frameL,None)
        kp2, des2 = sift.detectAndCompute(frameR,None)
        # FLANN parameters
        FLANN_INDEX_KDTREE = 1
        index_params = dict(algorithm = FLANN_INDEX_KDTREE, trees = 5)
        search_params = dict(checks=50)
        flann = cv.FlannBasedMatcher(index_params,search_params)
        matches = flann.knnMatch(des1,des2,k=2)
        pts1 = []
        pts2 = []
        # ratio test as per Lowe's paper
        for i,(m,n) in enumerate(matches):
            if m.distance < 0.8*n.distance:
                pts2.append(kp2[m.trainIdx].pt)
                pts1.append(kp1[m.queryIdx].pt)

        pts1 = np.int32(pts1)
        pts2 = np.int32(pts2)
        F, mask = cv.findFundamentalMat(pts1,pts2,cv.FM_LMEDS)
        # We select only inlier points
        pts1 = pts1[mask.ravel()==1]
        pts2 = pts2[mask.ravel()==1]
        # Find epilines corresponding to points in right image (second image) and
        # drawing its lines on left image
        lines1 = cv.computeCorrespondEpilines(pts2.reshape(-1,1,2), 2,F)
        lines1 = lines1.reshape(-1,3)
        img5, _ = draw_lines(frameL, frameR,lines1,pts1,pts2)
        # Find epilines corresponding to points in left image (first image) and
        # drawing its lines on right image
        lines2 = cv.computeCorrespondEpilines(pts1.reshape(-1,1,2), 1,F)
        lines2 = lines2.reshape(-1,3)
        img3, _ = draw_lines(frameR, frameL,lines2,pts2,pts1)
        return img5, img3

    def match(self, frameL, frameR, matcher: Matcher, metrics=True):
        """ Apply stereo SGBM.
        
        Args: 
            frameL: it's the left frame.
            frameR: it's the right frame.
            metrics: a boolean, if is true print by console the time of execution of correspondence step.

        Returns:
            left and right disparity maps and even matcher instance.
        """
        left_matcher = matcher.match()
        right_matcher = cv.ximgproc.createRightMatcher(left_matcher)
        if metrics:
            print('computing disparity...') 
            matching_time = time.time()
        left_disp = left_matcher.compute(frameL, frameR)
        right_disp = right_matcher.compute(frameR, frameL)
        if metrics:
            print("matching time: {} s".format(time.time() - matching_time))
        return left_disp, right_disp, left_matcher
         
    
    def post_process(self, frameL, left_disp, right_disp, matcher, lmbda=8000, sigma=1.5, metrics=True):
        """ Apply wls filter in disparity maps to smooth contours.
        
        Args: 
            frameL: it's the left frame.
            left_disp: it's the left disparity frame.
            right_disp: it's the right disparity frame.
            matcher: it's the matcher instance.
            lmbda: is a parameter defining the amount of regularization during filtering. Larger values force filtered disparity map edges to adhere more to source image edges. Typical value is 8000. Only valid in post processing step.
            sigma: is a parameter defining how sensitive the filtering process is to source image edges. Large values can lead to disparity leakage through low-contrast edges. Small values can make the filter too sensitive to noise and textures in the source image. Typical values range from 0.8 to 2.0. Only valid in post processing step
            metrics: if is true print by console the time of execution of post process step.

        Returns:
            Improved disparity map
        """
        wls_filter = cv.ximgproc.createDisparityWLSFilter(matcher)
        wls_filter.setLambda(lmbda);
        wls_filter.setSigmaColor(sigma);
        if metrics:
            postprocessing_time = time.time()
        filtered_disp = wls_filter.filter(left_disp, frameL, disparity_map_right=right_disp);
        if metrics:
            print("postprocessing time: {} s".format(time.time() - postprocessing_time))
        return filtered_disp

    def estimate_disparity_colormap(self, disparity):
        """ It converts disparity maps with a shape of (w, h, 1) to  (w, h, 3).

        Args: 
            disparity: a disparity map with a shape of (w, h, 1).

        Returns:
            disparity with color
        """
        return cv.applyColorMap((disparity).astype(np.uint8), cv.COLORMAP_JET)
    
    def estimate_depth_map(self, disparity, Q, frame, metrics=True):
        """ Convert disparity map to depth map.

        Args:
            frame: left or right frame.
            disparity: a disparity map with a shape of (w, h, 1)
            Q: a 4x4 array with the following structure, [[1 0   0          -cx     ][0 1   0          -cy     ][0 0   0           f      ][0 0 -1/Tx (cx - cx')/Tx ]] cx: is the principal point x in left image, cx': is the principal point x in right image, cy: is the principal point y in left image, f: is the focal lenth in left image, Tx: The x coordinate in Translation matrix.  
            metrics: a boolean, if is true print by console the time of execution of depth map step.

        Returns:
            reprojected points image and a depth map in RGB
        """
        if metrics:
            get_3D_points_time = time.time()
        disparity = disparity.astype(np.float32) / 16.0
        points = cv.reprojectImageTo3D(disparity, Q)
        colors = cv.cvtColor(frame, cv.COLOR_BGR2RGB)
        mask = disparity > disparity.min()
        out_points = points[mask]
        out_colors = colors[mask]
        if metrics:
            print("3D points time: {} s".format(time.time() - get_3D_points_time))
        return out_points, out_colors

    def estimate_3D_points(self, points, disparity, Q):
        """ Convert image plane points to homogeneous 3d points.

        Args:
            points: contains the (x, y) coordinates in image plane to convert to (X, Y, Z).
            disparity: a disparity map with a shape of (w, h, 1)
            Q: a 4x4 array with the following structure, [[1 0   0          -cx     ][0 1   0          -cy     ][0 0   0           f      ][0 0 -1/Tx (cx - cx')/Tx ]] cx: is the principal point x in left image, cx': is the principal point x in right image, cy: is the principal point y in left image, f: is the focal lenth in left image, Tx: The x coordinate in Translation matrix.  

        Returns:
            An array of points in 3D homogeneous coordinates (X, Y, Z, W).
        """
        points = np.array(points)
        disparity = np.array(disparity)
        homogeneous_3D_points = []
        for point in points:
            point_xy_disp = np.append(point, [disparity[point[1]][point[0]], 1], axis=0)
            projected_point = np.matmul(Q, point_xy_disp)
            homogeneous_3D_points.append(projected_point)
        return homogeneous_3D_points



from abc import ABCMeta, abstractmethod
from py2vision.models.yolov3_model import BuildYoloV3
from py2vision.models.yolov3_tiny_model import BuildYoloV3Tiny

class ModelManagerInterface(metaclass=ABCMeta):
    """
    An interface, which each method correspond with a network architecture.
    """
    @staticmethod
    @abstractmethod
    def build_yolov3():
        """Implements yolov3."""

    @staticmethod
    @abstractmethod
    def build_yolov3_tiny():
        """Implements yolov3 tiny"""


class ModelManager(ModelManagerInterface):
    """A selector of models that depends of which method is used.

    Attributes:
        model1: an instance of Yolov3Model
        model2: an instance of Yolov3TinyModel
    """
    def __init__(self):
        self.model1 = Yolov3Model()
        self.model2 = Yolov3TinyModel()
    
    def build_yolov3(self, backbone, num_class):
        """ Create the model to do YoloV3 based in darknet53.

        Args:
            backbone: an object with a backbone network.
            num_class: an integer with the quantity of classes.

        Returns:
            A list where the first one is used to predict large-sized objects, the second one is used to predict medium-sized objects, the third one is used to small objects and the last one is the input shape returned.
        """
        return self.model1.build(backbone, num_class)
    
    def build_yolov3_tiny(self, backbone, num_class):
        """ Create the model to do YoloV3 based in darknet19 tiny.

        Args:
            backbone: an object with a backbone network.
            num_class: an integer with the quantity of classes.

        Returns:
            A list where the first one is used to predict large-sized objects, the second one is used to predict medium-sized objects and the  last one is the input shape returned.
        """
        return self.model2.build(backbone, num_class)

class Yolov3Model:
    def build(self, backbone, num_class):
        return BuildYoloV3(backbone, num_class)

class Yolov3TinyModel:
    def build(self, backbone, num_class):
        return BuildYoloV3Tiny(backbone, num_class)


from __future__ import annotations
from abc import ABC, abstractmethod

import glob
import os
import xml.etree.ElementTree as ET


class Parser(ABC):
    """
    The Parser interface declares an `parse` method that should take the base AnnotationsFormat interface as an argument.
    """
    @abstractmethod
    def parse(self, anno: AnnotationsFormat):
        pass


class XmlParser(Parser):
    """
    Each Concrete Parser must implement the `parse` method in such a way that it calls the annotationsFormat's method corresponding to the Parser's class.
    """

    def parse(self, anno: AnnotationsFormat, xml_path, annotations_output_name, classes_output_name, image_path, work_dir=None, print_output=False):
        """
        This method convert annotations  from COCO or PASCAL VOC dataset in xml format to be compatible with an especific network model. Exporting a text file for annotations and a text file for classes names 

        Args:
            xml_path: a string with the full path of xml annotations.
            annotations_output_name: a string with the name of annotations file that will be generated.
            classes_output_name: a string with the name of classes file that will be generated.
            image_path: a full path where the images are saved.
            work_dir: a path where the annotations and classes files will be saved, if is None these will be saved in current directory.
            print_output: a boolean to print in console each annotation line
        """

        anno.visit_xml_parser(self, xml_path, annotations_output_name, classes_output_name, image_path, work_dir, print_output)

class AnnotationsFormat(ABC):
    """
    The AnnotationsFormat Interface declares a set of visiting methods that correspond to Parser classes. The signature of a visiting method allows the visitor to identify the exact class of the Parser that it's dealing with.
    """

    @abstractmethod
    def visit_xml_parser(self, element: XmlParser, xml_path, annotations_output_name, classes_output_name, image_path, work_dir=None, print_output=False):
        pass

class YoloV3AnnotationsFormat(AnnotationsFormat):
    """Get a group of xml annotations to transform in a .txt file compatible with YoloV3 dataset"""
    def visit_xml_parser(self, element ,xml_path, annotations_output_name, classes_output_name, image_path, work_dir=None, print_output=False):
        xmls = glob.glob(xml_path+'/*.xml')
        xmls = sorted(xmls)
        if len(xmls) == 0:
            raise FileNotFoundError("There isn't annotations in {}".format(xml_path))
        if work_dir is None:
            anno_file = os.getcwd()
            anno_file = os.path.join(anno_file, annotations_output_name)
            classes_file = os.getcwd()
            classes_file = os.path.join(classes_file, classes_output_name)
        else:
            anno_file = os.path.join(work_dir, annotations_output_name)
            classes_file = os.path.join(work_dir, classes_output_name)
        classes_names = []
        with open('{}.txt'.format(anno_file), 'w') as file:
            for xml_file in xmls:
                tree = ET.parse(open(xml_file))
                root = tree.getroot()
                image_name = root.find('filename').text
                img_path = image_path + '/' + image_name
                for i, obj in enumerate(root.iter('object')):
                    cls = obj.find('name').text
                    if cls not in classes_names:
                        classes_names.append(cls)
                    cls_id = classes_names.index(cls)
                    xmlbox = obj.find('bndbox')
                    object_params = (str(int(float(xmlbox.find('xmin').text)))+','
                            +str(int(float(xmlbox.find('ymin').text)))+','
                            +str(int(float(xmlbox.find('xmax').text)))+','
                            +str(int(float(xmlbox.find('ymax').text)))+','
                            +str(cls_id))
                    img_path += ' '+object_params
                if print_output:
                    print(img_path)
                file.write(img_path+'\n')
        
        with open('{}.txt'.format(classes_file), 'w') as file:
            for name in classes_names:
                file.write(str(name)+'\n')